@@ -1,14 +1,13 @@
-use crate::{
-    lexer::{Token, TokenIter},
-    ast::{Ast, Expr},
-    cmd,
-    cmd::Command,
-};
 use std::{
     collections::HashMap,
-    io::{BufRead, BufReader},
+    io::{self, BufRead, BufReader},
     fs::File,
 };
+use crate::{
+    lexer::{self, Token, TokenIter},
+    ast::{Ast, Expr},
+    cmd::{self, Command},
+};
 
 /// Our hand-written parser.
 /// Use with parse().
@@ -89,8 +88,8 @@ impl Parser {
     /// ```
     /// # use lambda_calc::{Parser, Ast, ast::Expr};
     /// let mut parser = Parser::new();
-    /// assert_eq!(parser.parse("K = lambda x y . x", None), None);
-    /// let redex = parser.parse("K a b", None);
+    /// assert_eq!(parser.parse("K = lambda x y . x", None), Ok(None));
+    /// let redex = parser.parse("K a b", None).unwrap();
     /// // we haven't asked for any beta reduction, so K has not been
     /// // substituted yet
     /// let expected = Ast::new(Expr::Redex(
@@ -102,7 +101,7 @@ impl Parser {
     /// )); // ((K a) b)
     /// assert_eq!(redex, Some(expected));
     /// let reduced = redex.unwrap().beta_reduce_quiet(&parser);
-    /// let expected = Ast::new(Expr::Var { name: "a".to_string(), is_free: true });
+    /// let expected = Some(Ast::new(Expr::Var { name: "a".to_string(), is_free: true }));
     /// assert_eq!(reduced, expected);
     /// ```
     ///
@@ -113,42 +112,105 @@ impl Parser {
     /// ```
     /// # use lambda_calc::Parser;
     /// let mut parser = Parser::new();
-    /// assert_eq!(parser.parse("foo = \\f -> f x", None), None);
+    /// assert_eq!(parser.parse("foo = \\f -> f x", None), Ok(None));
     /// let reduced = parser.parse("(\\x -> foo) a", None)
     ///     .unwrap()
+    ///     .unwrap()
     ///     .beta_reduce_quiet(&parser);
     /// // the "x" in foo's definition is never replaced with "a",
     /// // because the definition is only substituted later
     /// let expected = parser.parse("foo", None)
     ///     .unwrap()
+    ///     .unwrap()
     ///     .beta_reduce_quiet(&parser);
     /// assert_eq!(reduced, expected);
     /// ```
     ///
 
-    pub fn parse(&mut self, line: &str, file_info: Option<String>) -> Option<Ast> {
+    pub fn parse(&mut self, line: &str, file_info: Option<String>) -> Result<Option<Ast>, String> {
         let token_iter = TokenIter::new(&line);
         {
             let mut new_token_iter = token_iter.clone();
             if let Some(Token::Command(cmd)) = new_token_iter.next() {
                 self.run_command(cmd, new_token_iter.next());
-                return None;
+                return Ok(None);
             }
         }
         LineParser::new(token_iter, file_info).parse(self)
     }
 
-    /// Parse all lines in filename.
-    pub fn parse_file(&mut self, filename: &str) -> std::io::Result<()> {
-        let reader = BufReader::new(File::open(filename)?);
+    /// Parse all lines in the file named filename, or stdin if filename is None.
+    pub fn parse_file(&mut self, filename: Option<String>) -> Result<(), String> {
+        match filename {
+            None => {
+                let stdin = io::stdin();
+                let reader = stdin.lock();
+                self.parse_file_with_reader(reader, "stdin")?;
+            },
+            Some(s) => {
+                let file = match File::open(&s) {
+                    Ok(file) => file,
+                    Err(e) => return Err(e.to_string()),
+                };
+                let reader = BufReader::new(file);
+                self.parse_file_with_reader(reader, &format!("file '{}'", s))?;
+            },
+        }
+        Ok(())
+    }
 
-        for (line_num, line) in reader.lines().enumerate() {
-            let file_info = format!("file '{}', line {}: ", filename, line_num + 1);
-            if let Some(ast) = self.parse(&line?, Some(file_info)) {
+    fn parse_file_with_reader<R: BufRead>(&mut self,
+                                          reader: R,
+                                          file_info_prefix: &str) -> Result<(), String> {
+
+        let mut line_num_beg = 0;
+        let mut line_num_end;
+        let mut lines = reader.lines();
+        'read_expr: loop {
+            let mut expr = match lines.next() {
+                None => break 'read_expr,
+                Some(line) => {
+                    line_num_beg += 1;
+                    match line {
+                        Ok(l) => l,
+                        Err(e) => return Err(e.to_string()),
+                    }
+                },
+            };
+
+            line_num_end = line_num_beg;
+
+            while lexer::strip_whitespace_and_line_cont(&mut expr) {
+                let line = match lines.next() {
+                    None => return Err(format!("{}: line continuation not allowed in file's last line",
+                                               file_info_prefix)),
+                    Some(line) => {
+                        line_num_end += 1;
+                        match line {
+                            Ok(l) => l,
+                            Err(e) => return Err(e.to_string()),
+                        }
+                    }
+                };
+                expr.push_str(&line);
+            }
+
+            let file_info = if line_num_end > line_num_beg {
+                format!("{}, lines {}-{}: ",
+                        file_info_prefix, line_num_beg, line_num_end)
+            } else {
+                format!("{}, line {}: ", file_info_prefix, line_num_beg)
+            };
+
+            if let Some(ast) = self.parse(&expr, Some(file_info))? {
                 if self.non_interactive_mode {
-                    println!("{:#}", ast.beta_reduce_quiet(&self));
+                    if let Some(ast) = ast.beta_reduce_quiet(&self) {
+                        println!("{:#}", ast);
+                    }
                 }
             }
+
+            line_num_beg = line_num_end;
         }
         Ok(())
     }
@@ -195,14 +257,16 @@ impl Parser {
                     return;
                 },
                 Some(Token::Id(filename)) => {
-                    if let Err(err) = self.parse_file(filename) {
+                    if let Err(err) = self.parse_file(Some(filename.to_string())) {
                         eprintln!("error parsing file {}: '{}'", filename, err);
                     }
                 },
                 _ => panic!("lexer should have returned an Id as the argument to ':load'"),
             },
             Command::Define => panic!("Command::Define should never be returned by the lexer!"),
-            Command::Pause => handle_bool_cmd(&mut self.pause, arg),
+            Command::Pause => if !self.non_interactive_mode {
+                handle_bool_cmd(&mut self.pause, arg);
+            },
             Command::Step => handle_bool_cmd(&mut self.step, arg),
             Command::CountSteps => handle_bool_cmd(&mut self.count_steps, arg),
         }
@@ -229,20 +293,18 @@ impl<'a, I> LineParser<'a, I>
         }
     }
 
-    fn parse(&mut self, parser: &mut Parser) -> Option<Ast> {
+    fn parse(&mut self, parser: &mut Parser) -> Result<Option<Ast>, String> {
         if let None = self.token_iter.clone().next() {
-            return None; // empty line, ignore.
+            return Ok(None); // empty line, ignore.
         }
         let is_def = self.check_is_def();
-        if !self.sanity_checks(is_def) {
-            return None;
-        }
+        self.sanity_checks(is_def)?;
         if is_def {
             self.parse_def(parser);
-            None
+            Ok(None)
         } else {
             let ast = self.parse_ast(Vec::new())?;
-            Some(ast)
+            Ok(ast)
         }
     }
 
@@ -256,14 +318,16 @@ impl<'a, I> LineParser<'a, I>
             _ => panic!("expected definition, but 2nd token is not '=' or ':='"),
         }
         match self.parse_ast(Vec::new()) {
-            None => self.print_err_msg("a definition can't bind to an empty expression"),
-            Some(ast) => match ast.expr_ref() {
-                Expr::LambdaTerm { var_name: _, body: _ } => {
-                    parser.insert_symbol(name,  ast);
+            Err(e) => self.print_syntax_err(&e),
+            Ok(None) => self.print_syntax_err("a definition can't bind to an empty expression"),
+            Ok(Some(ast)) => match ast.expr_ref() {
+                Expr::LambdaTerm { .. } => {
+                    parser.insert_symbol(name, ast);
                 },
                 Expr::Redex(_,_) => {
-                    let non_redex = ast.beta_reduce_quiet(parser);
-                    parser.insert_symbol(name, non_redex);
+                    if let Some(non_redex) = ast.beta_reduce_quiet(parser) {
+                        parser.insert_symbol(name, non_redex);
+                    }
                 },
                 Expr::Var{name: var_name, is_free, } => {
                     assert!(is_free);
@@ -272,7 +336,7 @@ impl<'a, I> LineParser<'a, I>
                         // bind to the symbol's definition directly
                         parser.insert_symbol(name, ast);
                     } else {
-                        self.print_err_msg("a definition can't bind to a single variable (unless it also has a definition)");
+                        self.print_syntax_err("a definition can't bind to a single variable (unless it also has a definition)");
                     }
                 }
             }
@@ -281,36 +345,32 @@ impl<'a, I> LineParser<'a, I>
 
     // Our grammar is not LR(1) (and may not be context-free either),
     // so we parse by hand.
-    fn parse_ast(&mut self, mut queue: Vec<Ast>) -> Option<Ast> {
+    fn parse_ast(&mut self, mut queue: Vec<Ast>) -> Result<Option<Ast>, String> {
         loop {
             match self.token_iter.next() {
                 None | Some(Token::CloseParen) => {
-                    return finalize_redex(queue);
+                    return Ok(finalize_redex(queue));
                 },
                 Some(Token::OpenParen) => {
-                    queue.push(
-                        self.parse_ast(Vec::new())
-                        .expect("null expression after open paren")
-                    );
+                    let ast = match self.parse_ast(Vec::new())? {
+                        None => return Err("null expression after open paren".to_string()),
+                        Some(x) => x,
+                    };
+                    queue.push(ast);
                 },
                 Some(Token::Id(s)) => {
                     let is_free = !self.lambda_vars.iter().any(|x| x == s);
                     queue.push(Ast::new(Expr::Var { name: s.to_string(), is_free, }));
                 },
                 Some(Token::Lambda) => {
-                    let (ast, var_count) = self.parse_lambda();
+                    let (ast, var_count) = self.parse_lambda(false)?;
                     queue.push(ast);
                     for _ in 0..var_count {
                         self.lambda_vars.pop();
                     }
-                    return finalize_redex(queue);
-                },
-                t @ _ => {
-                    panic!(format!(
-                        "syntax error, token '{:?}'; should've been handled by sanity_checks()!",
-                        t
-                    ));
+                    return Ok(finalize_redex(queue));
                 },
+                Some(t) => return Err(format!("unexpected token '{:?}'", t)),
             }
         }
     }
@@ -324,33 +384,36 @@ impl<'a, I> LineParser<'a, I>
     // The number returned is the number of lambda variables pushed into
     // self.lambda_vars.
     //
-    fn parse_lambda(&mut self) -> (Ast, i32) {
+    fn parse_lambda(&mut self, var_strict: bool) -> Result<(Ast, i32), String> {
         match self.token_iter.next() {
             Some(Token::Id(name)) => {
                 if self.lambda_vars.iter().any(|x| x == name) {
-                    // Ideally we should do something less drastic than panicking,
-                    // e.g. returning None, but that could lead to a non-None
-                    // inconsistent expression, which is even worse.
-                    //
-                    panic!("outer lambda variable cannot appear again as an inner lambda variable");
+                    return Err(format!("variable name '{}' is already in use by an inner lambda term", name));
                 }
                 self.lambda_vars.push(name.to_string());
 
-                let (lambda_body, var_count) = self.parse_lambda();
+                let (lambda_body, var_count) = self.parse_lambda(false)?;
                 let ast = Ast::new(Expr::LambdaTerm {
                     var_name: name.to_string(),
+                    var_strict,
                     body: Box::new(lambda_body),
                 });
 
-                (ast, var_count + 1)
+                Ok((ast, var_count + 1))
+            },
+            Some(Token::Strict) => {
+                self.parse_lambda(true)
             },
             Some(Token::Gives) => {
                 // finalize lambda term by returning its body.
-                let ast = self.parse_ast(Vec::new())
-                    .expect("lambda term can't have empty body");
-                (ast, 0)
+                let ast = match self.parse_ast(Vec::new())? {
+                    None => return Err("lambda term can't have empty body".to_string()),
+                    Some(x) => x,
+                };
+                Ok((ast, 0))
             },
-            _ => panic!("this token is not allowed in the head of a lambda term."),
+            Some(t) => Err(format!("token {:?} is not allowed in lambda term's head", t)),
+            None => Err(format!("unfinished lambda term head")),
         }
     }
 
@@ -364,44 +427,46 @@ impl<'a, I> LineParser<'a, I>
         false
     }
 
-    fn sanity_checks(&self, is_def: bool) -> bool {
+    fn sanity_checks(&self, is_def: bool) -> Result<(), String> {
         let mut paren_count:i32 = 0;
         let token_iter = self.token_iter.clone();
 
         for (i, token) in token_iter.enumerate() {
             match token {
-                Token::Invalid => {
-                    self.print_err_msg(&format!("token number {} is invalid", i));
-                    return false;
+                Token::Invalid(c) => {
+                    return Err(self.prepend_file_info(format!("token {:?} is invalid", c)));
                 },
                 Token::OpenParen => paren_count += 1,
                 Token::CloseParen => paren_count -= 1,
                 Token::Def => {
                     if !is_def {
-                        self.print_err_msg(
-                            &format!("wrong syntax for definition, should be <var> = <expr> (token {})", i)
-                        );
-                        return false;
+                        return Err(self.prepend_file_info(format!("wrong syntax for definition, should be <var> = <expr> (token {})", i)));
                     }
                 },
                 _ => {},
             }
         }
         if paren_count > 0 {
-            self.print_err_msg(&format!("{} unclosed parentheses", paren_count));
-            return false;
+            return Err(self.prepend_file_info(format!("{} unclosed parentheses", paren_count)));
         } else if paren_count < 0 {
-            self.print_err_msg(&format!("{} extra closing parentheses", -paren_count));
-            return false;
+            return Err(self.prepend_file_info(format!("{} extra closing parentheses", -paren_count)));
         }
-        true
+        Ok(())
     }
 
-    fn print_err_msg(&self, s: &str) {
+    fn prepend_file_info(&self, s: String) -> String {
+        if let Some(fs) = &self.file_info {
+            format!("{} {}", fs, s)
+        } else {
+            s
+        }
+    }
+
+    fn print_syntax_err(&self, s: &str) {
         if let Some(fs) = &self.file_info {
             eprint!("{}", fs);
         }
-        eprintln!("{}", s);
+        eprintln!("syntax error: {}", s);
     }
 }
 
@@ -451,9 +516,10 @@ mod tests {
             is_free: false,
         })
     }
-    fn lambda_no_box(var_name: &str, body: Box<Ast>) -> Ast {
+    fn lambda_no_box(var_name: &str, var_strict: bool, body: Box<Ast>) -> Ast {
         Ast::new(Expr::LambdaTerm {
             var_name: var_name.to_string(),
+            var_strict,
             body,
         })
     }
@@ -469,8 +535,8 @@ mod tests {
     fn bound_var(name: &str) -> Box<Ast> {
         Box::new(bound_var_no_box(name))
     }
-    fn lambda(var_name: &str, body: Box<Ast>) -> Box<Ast> {
-        Box::new(lambda_no_box(var_name, body))
+    fn lambda(var_name: &str, var_strict: bool, body: Box<Ast>) -> Box<Ast> {
+        Box::new(lambda_no_box(var_name, var_strict, body))
     }
     fn redex(left: Box<Ast>, right: Box<Ast>) -> Box<Ast> {
         Box::new(redex_no_box(left, right))
@@ -480,14 +546,14 @@ mod tests {
     fn empty_expr() {
         let mut parser = Parser::new();
         let expr = LineParser::new(vec![].into_iter(), None).parse(&mut parser);
-        assert_eq!(None, expr);
+        assert_eq!(Ok(None), expr);
     }
 
     // skeleton for non-empty expressions' tests (not definitions)
     fn expr_test<'a>(tokens: Vec<Token<'a>>, expected: Ast) {
         let mut parser = Parser::new();
         let ast = LineParser::new(tokens.into_iter(), None).parse(&mut parser);
-        assert_eq!(Some(expected), ast);
+        assert_eq!(Ok(Some(expected)), ast);
     }
 
     #[test]
@@ -577,7 +643,7 @@ mod tests {
         ]; // lambda x . x
         expr_test(
             tokens,
-            lambda_no_box("x", bound_var("x"))
+            lambda_no_box("x", false, bound_var("x"))
         );
     }
 
@@ -592,7 +658,9 @@ mod tests {
 
         expr_test(
             tokens,
-            lambda_no_box("x", lambda("y", lambda("z", free_var("a"))))
+            lambda_no_box("x", false,
+                          lambda("y", false,
+                                 lambda("z", false, free_var("a"))))
         );
     }
 
@@ -624,7 +692,7 @@ mod tests {
         ]; // lambda x . (x a)
         expr_test(
             tokens,
-            lambda_no_box("x", redex(bound_var("x"), free_var("a")))
+            lambda_no_box("x", false, redex(bound_var("x"), free_var("a")))
         );
     }
 
@@ -645,9 +713,9 @@ mod tests {
         expr_test(
             tokens,
             lambda_no_box(
-                "x",
-                lambda("y",
-                       lambda("z",
+                "x", false,
+                lambda("y", false,
+                       lambda("z", false,
                               redex(
                                   redex(bound_var("x"), bound_var("y")),
                                   bound_var("z")
@@ -678,13 +746,13 @@ mod tests {
         expr_test(
             tokens,
             lambda_no_box(
-                "x",
+                "x", false,
                 redex(
                     bound_var("x"),
-                    lambda("y",
+                    lambda("y", false,
                            redex(
                                bound_var("x"),
-                               lambda("z", bound_var("x"))
+                               lambda("z", false, bound_var("x"))
                            )
                     )
                 )
@@ -702,7 +770,7 @@ mod tests {
             tokens,
             redex_no_box(
                 free_var("a"),
-                lambda("x", bound_var("x"))
+                lambda("x", false, bound_var("x"))
             ),
         );
     }
@@ -761,7 +829,7 @@ mod tests {
             tokens,
             redex_no_box(
                 redex(
-                    lambda("x", bound_var("x")),
+                    lambda("x", false, bound_var("x")),
                     free_var("a"),
                 ),
                 free_var("b"),
@@ -827,7 +895,7 @@ mod tests {
         expr_test(
             tokens,
             lambda_no_box(
-                "x",
+                "x", false,
                 redex(
                     redex(free_var("a"), free_var("b")),
                     free_var("c")
@@ -847,7 +915,7 @@ mod tests {
         expr_test(
             tokens,
             redex_no_box(
-                lambda("x", bound_var("x")),
+                lambda("x", false, bound_var("x")),
                 free_var("x")
             ),
         );
@@ -870,8 +938,8 @@ mod tests {
             tokens,
             redex_no_box(
                 redex(
-                    lambda("x",
-                           lambda("y",
+                    lambda("x", false,
+                           lambda("y", false,
                                   redex(bound_var("x"), bound_var("y"))
                            )
                     ),
@@ -896,7 +964,7 @@ mod tests {
             redex_no_box(
                 redex(
                     free_var("a"),
-                    lambda("x", bound_var("x")),
+                    lambda("x", false, bound_var("x")),
                 ),
                 free_var("b")
             ),
@@ -917,7 +985,7 @@ mod tests {
             redex_no_box(
                 redex(
                     free_var("a"),
-                    lambda("x", bound_var("x")),
+                    lambda("x", false, bound_var("x")),
                 ),
                 free_var("b"),
             ),
@@ -941,7 +1009,7 @@ mod tests {
                 redex(
                     redex(
                         redex(free_var("a"), free_var("b")),
-                        lambda("x", bound_var("x")),
+                        lambda("x", false, bound_var("x")),
                     ),
                     free_var("c"),
                 ),
@@ -967,10 +1035,10 @@ mod tests {
             tokens,
             redex_no_box(
                 redex(
-                    lambda("x", bound_var("x")),
-                    lambda("x", bound_var("x")),
+                    lambda("x", false, bound_var("x")),
+                    lambda("x", false, bound_var("x")),
                 ),
-                lambda("x", bound_var("x")),
+                lambda("x", false, bound_var("x")),
             ),
         );
     }
@@ -995,7 +1063,7 @@ mod tests {
                 redex(
                     redex(
                         free_var("a"),
-                        lambda("x", bound_var("x")),
+                        lambda("x", false, bound_var("x")),
                     ),
                     free_var("b"),
                 ),
@@ -1018,19 +1086,91 @@ mod tests {
         expr_test(
             tokens,
             lambda_no_box(
-                "f",
+                "f", false,
                 redex(
-                    lambda("x", bound_var("f")),
-                    lambda("x", bound_var("f")),
+                    lambda("x", false, bound_var("f")),
+                    lambda("x", false, bound_var("f")),
                 ),
             ),
         );
     }
 
+    #[test]
+    fn single_strict_var_singlevar_lambda() {
+        let tokens = vec![
+            Token::Lambda, Token::Strict, Token::Id("x"),
+            Token::Gives,
+            Token::Id("x")
+        ]; // \!x -> x
+        expr_test(
+            tokens,
+            lambda_no_box("x", true, bound_var("x"))
+        );
+    }
+
+    #[test]
+    fn single_strict_var_multivar_lambda1() {
+        let tokens = vec![
+            Token::Lambda, Token::Strict, Token::Id("x"), Token::Id("y"),
+            Token::Gives,
+            Token::Id("x"), Token::Id("y")
+        ]; // \!x y -> x y
+        expr_test(
+            tokens,
+            lambda_no_box(
+                "x", true,
+                lambda(
+                    "y", false,
+                    redex(bound_var("x"), bound_var("y"))
+                )
+            )
+        );
+    }
+
+    #[test]
+    fn single_strict_var_multivar_lambda2() {
+        let tokens = vec![
+            Token::Lambda, Token::Id("x"), Token::Strict, Token::Id("y"),
+            Token::Gives,
+            Token::Id("x"), Token::Id("y")
+        ]; // \x !y -> x y
+        expr_test(
+            tokens,
+            lambda_no_box(
+                "x", false,
+                lambda(
+                    "y", true,
+                    redex(bound_var("x"), bound_var("y"))
+                )
+            )
+        );
+    }
+
+    #[test]
+    fn multi_strict_var_multivar_lambda() {
+        let tokens = vec![
+            Token::Lambda, Token::Strict, Token::Id("x"),
+            Token::Id("y"),
+            Token::Strict, Token::Id("z"),
+            Token::Gives,
+            Token::Id("z")
+        ]; // \!x y !z -> z
+        expr_test(
+            tokens,
+            lambda_no_box(
+                "x", true,
+                lambda(
+                    "y", false,
+                    lambda("z", true, bound_var("z"))
+                )
+            )
+        );
+    }
+
     #[test]
     fn insert_get_symbol() {
         let mut parser = Parser::new();
-        let ast = lambda_no_box("x", bound_var("x"));
+        let ast = lambda_no_box("x", false, bound_var("x"));
         parser.insert_symbol("I", ast.clone());
         assert_eq!(parser.get_symbol("I"), Some(&ast));
     }
@@ -1039,7 +1179,7 @@ mod tests {
     fn def_test<'a>(name: &str, def_tokens: Vec<Token<'a>>, expected: Option<&Ast>) {
         let mut parser = Parser::new();
         assert_eq!(
-            None,
+            Ok(None),
             LineParser::new(def_tokens.into_iter(), None).parse(&mut parser)
         );
         let ast = parser.get_symbol(name);
@@ -1052,7 +1192,7 @@ mod tests {
             Token::Id("I"), Token::Def,
             Token::Lambda, Token::Id("x"), Token::Gives, Token::Id("x")
         ];
-        let expected = lambda_no_box("x", bound_var("x"));
+        let expected = lambda_no_box("x", false, bound_var("x"));
         def_test("I", def_tokens, Some(&expected));
     }
 
@@ -1075,16 +1215,16 @@ mod tests {
         ];
         let mut parser = Parser::new();
         assert_eq!(
-            None,
+            Ok(None),
             LineParser::new(def1.into_iter(), None).parse(&mut parser)
         );
         assert_eq!(
-            None,
+            Ok(None),
             LineParser::new(def2.into_iter(), None).parse(&mut parser)
         );
         assert_eq!(
             parser.get_symbol("g"),
-            Some(&lambda_no_box("x", bound_var("x"))),
+            Some(&lambda_no_box("x", false, bound_var("x"))),
         );
     }
 }
